/* Include the controller definition */

#include "epuck_reality_gap.h"
/* Function definitions for XML parsing */
#include <argos3/core/utility/configuration/argos_configuration.h>


/****************************************/
/****************************************/

CEpuckRealityGap::CEpuckRealityGap() :
    m_pcWheelAct(NULL),
    m_pcProxSens(NULL),
    m_pcRGBLedsAct(NULL),
    m_pcRNG(NULL),
    m_bIsRED(false),
    m_pcTcpConnection(NULL),
    m_bCanStart(false),
    m_bIsExperimentFinished(false)
   {}

/****************************************/
/****************************************/

void CEpuckRealityGap::Init(TConfigurationNode& t_node) {
    m_pcWheelAct = GetActuator<CCI_EPuckWheelsActuator>("epuck_wheels");
    m_pcProxSens = GetSensor<CCI_EPuckProximitySensor>("epuck_proximity");
    m_pcRGBLedsAct = GetActuator<CCI_EPuckRGBLEDsActuator>("epuck_rgb_leds");
    
    try {
        TConfigurationNode implementation = GetNode(t_node, "implementation");
        GetNodeAttribute(implementation, "real_robots", m_bIsRunningRealRobots);
        
        TConfigurationNode tcp = GetNode(t_node, "tcpconnection");
        GetNodeAttribute(tcp, "host", m_sHost);
    } catch(CARGoSException& ex) {
        THROW_ARGOSEXCEPTION_NESTED("Error initializing controller parameters.", ex);
    }

    m_unNbCs = 0;
    m_pcRNG = CRandom::CreateRNG("argos");
    m_bOAGoRight = false;

    m_bIsRED = false;
    m_unREDLEDsCpt = 0;
    m_unNbrTurnedRed = 0;

    m_bChangeDirection = false;
    m_bChangeDirectionRight = true;
    m_unChangeDirectionCpt = 15;
    m_unWaitBeforeReturnRed = 60;
    m_bCanTurnRed = true;
    m_bWaitForExperimentToStart = true;
    m_unWaitBeforeCheckingObstacle = 0;
    m_unCScpt = 0;
    
    LOG << "Connection to server" << std::endl;
   	m_pcTcpConnection = new TCPConnection(m_sHost, 4321);
    m_pcTcpConnection->Start();

}


/****************************************/
/****************************************/

void CEpuckRealityGap::ControlStep() {

	ReadIncomingTCPMessages();

    if (m_bWaitForExperimentToStart || m_bIsExperimentFinished ) {
     	m_pcWheelAct->SetLinearVelocity(0.0, 0.0);
        return;
    }
    
    TurnLEDsOff();
    if (ObstacleDetected()) {
        m_bObstacleAvoidanceState = true;
        if (m_unWaitBeforeCheckingObstacle > 0) {
            m_unWaitBeforeCheckingObstacle--;

            return;
        }
        ObstacleAvoidance();
        return;
    }

    if (m_bObstacleAvoidanceState && !ObstacleDetected()) {
        m_bObstacleAvoidanceState = false;
    }
    if (!m_bObstacleAvoidanceState && !m_bChangeDirection) {
        m_pcWheelAct->SetLinearVelocity(10.0, 10.0);

        if (m_pcRNG->Bernoulli(0.02)) {
            m_bChangeDirection = false;
        } else {
            m_bChangeDirection = false;
        }
    }
    
    if (m_bChangeDirection) {
        if (m_unChangeDirectionCpt > 0) {
            if (m_bChangeDirectionRight) {
                m_pcWheelAct->SetLinearVelocity(10.0, 5.0);
            }
            else {
                m_pcWheelAct->SetLinearVelocity(5.0, 10.0);
            }
            --m_unChangeDirectionCpt;
        }
        else {
            m_bChangeDirection = false;
            m_unChangeDirectionCpt = m_pcRNG->Uniform(CRange<UInt32>(20,30));
            m_bChangeDirectionRight = m_pcRNG->Bernoulli(0.5);
            m_unWaitBeforeCheckingObstacle = 15;
        }
        
    }   
}


/****************************************/
/****************************************/
void CEpuckRealityGap::TurnRed() {
 
    m_pcRGBLedsAct->SetColors(CColor::RED);
    m_bIsRED = true;
}


/****************************************/
/****************************************/
void CEpuckRealityGap::TurnLEDsOff() {
    if (m_bIsRED && m_unREDLEDsCpt < 20) {
        m_unREDLEDsCpt++;
    }
    else if (m_bIsRED && m_unREDLEDsCpt >= 20) {
        m_pcRGBLedsAct->SetColors(CColor::BLACK);
        m_unREDLEDsCpt = 0;
        m_bIsRED = false;
    }
}

/****************************************/
/****************************************/
void CEpuckRealityGap::ObstacleAvoidance() {
    Real velocity = 3.0f;

	 if (m_bOAGoRight) {
	       m_pcWheelAct->SetLinearVelocity(velocity, -velocity);
	    } else {
	    	m_pcWheelAct->SetLinearVelocity(-velocity, velocity);
	    }
}

/****************************************/
/****************************************/

bool CEpuckRealityGap::ObstacleDetected() {
	
	//Real fProximityThreshold = .001;
	// Used in real experiment (experiments done in may 2015))
	Real fProximityThreshold = .1;

	const CCI_EPuckProximitySensor::TReadings& tReadings = m_pcProxSens->GetReadings();

	if (tReadings[0].Value > fProximityThreshold && tReadings[7].Value > fProximityThreshold) {
		if (m_pcRNG->Bernoulli(0.5)) {
			m_bOAGoRight = true;
		}
		else {
			m_bOAGoRight = false;
		}
        return true;
	}

	else if (tReadings[0].Value > fProximityThreshold && tReadings[1].Value > fProximityThreshold) {
		m_bOAGoRight = true;

	}
	else if (tReadings[7].Value > fProximityThreshold && tReadings[6].Value > fProximityThreshold) {
		m_bOAGoRight = false;
	}

	else if (tReadings[0].Value > fProximityThreshold 
            && tReadings[1].Value <= fProximityThreshold && tReadings[6].Value <= fProximityThreshold
			&& tReadings[7].Value <= fProximityThreshold) {
		m_bOAGoRight =true;
		return true;
	}

	else if (tReadings[7].Value > fProximityThreshold 
            && tReadings[1].Value <= fProximityThreshold && tReadings[6].Value <= fProximityThreshold
			&& tReadings[0].Value <= fProximityThreshold) {
		m_bOAGoRight =false;
		return true;
	}
	else if (tReadings[1].Value > fProximityThreshold 
            && tReadings[0].Value <= fProximityThreshold && tReadings[6].Value <= fProximityThreshold
			&& tReadings[7].Value <= fProximityThreshold) {
		m_bOAGoRight =true;
		return true;
	}
	else if (tReadings[6].Value > fProximityThreshold 
            && tReadings[1].Value <= fProximityThreshold && tReadings[0].Value <= fProximityThreshold
			&& tReadings[7].Value <= fProximityThreshold) {
		m_bOAGoRight =false;
		return true;
	}

	else {
		return false;
	}

	return true;


}


/****************************************/
/****************************************/
void CEpuckRealityGap::ReadIncomingTCPMessages() {
    std::string sMsg = m_pcTcpConnection->GetLatestMessage();

    if ((sMsg == "r" && m_bIsRunningRealRobots) ||(sMsg == "a" && !m_bIsRunningRealRobots)) {
          m_bWaitForExperimentToStart = false;
          LOG << "Received message start" << std::endl;
    } 
    else if (sMsg.substr(0, 5) == "epuck") {
    	LOG << "receive epuck: " << sMsg << " my id: " << GetId() << std::endl;
        if (sMsg == GetId()) {
        	TurnRed();
        }
    } 
    else if (sMsg == "stop") {
        m_bIsExperimentFinished = true;
    }
}

/****************************************/
/****************************************/



REGISTER_CONTROLLER(CEpuckRealityGap, "epuck_reality_gap")
